import os
import re
import shutil
import configparser
import pyodbc
import backoff

# pylint: disable=no-member

def _write_userprefs(host, port):
    """Creates or updates the ~/.iSeriesAccess/cwb_userprefs.ini file to
    specify a "Port lookup mode", which controls how the driver determines
    which port to use when connecting."""
    # This file and its values were found by chris@stitchdata.com using strace,
    # educated guesses, and hints from random docs on the internet. At the time
    # of this writing, I could not find any specific documentation on this file
    # or its options.
    config = configparser.ConfigParser()
    fname = os.path.expanduser("~/.iSeriesAccess/cwb_userprefs.ini")
    if not os.path.exists(fname):
        os.makedirs(os.path.dirname(fname), 0o700, exist_ok=True)
    else:
        config.read(fname)
    section = (r"CWB_CURRUSER\Software\IBM\Client Access Express\CurrentVersion"
               r"\Environments\My Connections\{}\Communication".format(host))
    if not section in config:
        config.add_section(section)
    key = "Port lookup mode"
    port_mode = "attr_dwd:0x00000001" if port else "attr_dwd:0x00000002"
    if config[section].get(key) != port_mode:
        config[section][key] = port_mode
        with open(fname, "w") as f:
            config.write(f)


def _write_port_to_services(port):
    """Modifies the /etc/services file to specify a port for connecting to
    DB2."""
    with open("/etc/services", "r+") as f:
        lines = f.readlines()
        f.seek(0)
        for line in lines:
            # as-database was found in
            # https://www.ibm.com/support/knowledgecenter/en/ssw_ibm_i_71/rzaii/rzaiiservicesandports.htm
            if not re.match(r"^as-database\s+", line):
                f.write(line)
        f.truncate()
        f.write("as-database {}/tcp # tap-db2\n".format(port))


def setup_port_configuration(config):
    host = config["host"]
    port = config.get("port")
    _write_userprefs(host, port)
    if port:
        _write_port_to_services(port)


@backoff.on_exception(backoff.expo,
                      (pyodbc.OperationalError),
                      max_tries=5,
                      factor=2)
def connection(config):
    # Docs on keywords this driver accepts:
    # https://www.ibm.com/support/knowledgecenter/ssw_ibm_i_71/rzaik/rzaikconnstrkeywordsgeneralprop.htm
    return pyodbc.connect(
        driver="{iSeries Access ODBC Driver 64-bit}",
        system=config["host"],
        uid=config["user"],
        pwd=config["password"])


class get_cursor(object):
    def __init__(self, config):
        self.conn = connection(config)
        self.cur = self.conn.cursor()

    def __enter__(self):
        return self.cur

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.cur.close()
        self.conn.close()



#!/usr/bin/env python3
import singer
from singer import utils
from singer.catalog import Catalog
from . import resolve, sync, discovery, common

REQUIRED_CONFIG_KEYS = ["host", "user", "password"]
LOGGER = singer.get_logger()


def do_sync(args, input_catalog):
    state = resolve.build_state(args.state, input_catalog)
    catalog = resolve.resolve_catalog(input_catalog, input_catalog, state)
    sync.sync(args.config, state, catalog)


def main_impl():
    args = utils.parse_args(REQUIRED_CONFIG_KEYS)
    common.setup_port_configuration(args.config)
    if args.discover:
        discovery.discover(args.config).dump()
        print()
    elif args.catalog:
        do_sync(args, args.catalog)
    elif args.properties:
        do_sync(args, Catalog.from_dict(args.properties))
    else:
        LOGGER.info("No properties were selected")


def main():
    try:
        main_impl()
    except Exception as exc:
        LOGGER.critical("unknown top-level tap exception", exc_info=exc)
        raise



"""Contains functionality for converting DB2 catalog results into
Singer schemas."""
from singer.catalog import Schema

# https://www.ibm.com/support/knowledgecenter/ssw_ibm_i_71/db2/rbafzch2num.htm
BYTES_FOR_INTEGER_TYPE = {
    "smallint": 2,
    "integer": 4,
    "bigint": 8,
}
FLOAT_TYPES = {
    "float",
    "decfloat",
}
DECIMAL_TYPES = {
    "decimal",
    "numeric",
}

# https://www.ibm.com/support/knowledgecenter/ssw_ibm_i_71/db2/rbafzcharstrg.htm
STRING_TYPES = {
    "char",
    "varchar",
}

# https://www.ibm.com/support/knowledgecenter/ssw_ibm_i_71/db2/rbafzch2datetime.htm
DATETIME_TYPES = {
    "date",
    "timestmp",
    "time",
}

VALID_REPLICATION_KEY_TYPES = {
    "smallint",
    "integer",
    "bigint",
    "decimal",
    "numeric",
    "date",
    "timestmp",
}

# Parent article for data types:
# https://www.ibm.com/support/knowledgecenter/ssw_ibm_i_71/db2/rbafzch2data.htm
def _for_column(col, pk_columns):
    data_type = col.data_type.lower()
    inclusion = "available"
    if col.column_name.lower() in [x.lower() for x in pk_columns]:
        inclusion = "automatic"
    result = Schema(inclusion=inclusion)
    if data_type in BYTES_FOR_INTEGER_TYPE:
        result.type = ["null", "integer"]
        bits = BYTES_FOR_INTEGER_TYPE[data_type] * 8
        result.minimum = 0 - 2 ** (bits - 1)
        result.maximum = 2 ** (bits - 1) - 1
    elif data_type in FLOAT_TYPES:
        result.type = ["null", "number"]
    elif data_type in DECIMAL_TYPES:
        result.type = ["null", "number"]
        result.exclusiveMaximum = True
        result.maximum = 10 ** (col.numeric_precision - col.numeric_scale)
        result.exclusiveMinimum = True
        result.minimum = -10 ** (col.numeric_precision - col.numeric_scale)
        result.multipleOf = 10 ** (0 - col.numeric_scale)
    elif data_type in STRING_TYPES:
        result.type = ["null", "string"]
        if col.character_maximum_length > 0:
            result.maxLength = col.character_maximum_length
    elif data_type in DATETIME_TYPES:
        result.type = ["null", "string"]
        result.format = "date-time"
    else:
        err = "Unsupported data type {}".format(data_type)
        result = Schema(None, inclusion="unsupported", description=err)
    return result


def generate(columns, pk_columns):
    properties = {c.column_name: _for_column(c, pk_columns) for c in columns}
    return Schema(type="object", properties=properties)


def valid_replication_keys(columns):
    return [c.column_name for c in columns
            if c.data_type.lower() in VALID_REPLICATION_KEY_TYPES]



#!/usr/bin/env python3
from setuptools import setup

setup(
    name="tap-db2",
    version="0.2.6",
    description="Singer.io tap for extracting data from DB2",
    author="Stitch",
    url="http://singer.io",
    classifiers=["Programming Language :: Python :: 3 :: Only"],
    install_requires=[
        "singer-python==5.0.3",
        "pyodbc>4,<5",
    ],
    entry_points="""
    [console_scripts]
    tap-db2=tap_db2:main
    """,
    packages=["tap_db2", "tap_db2.discovery"],
    include_package_data=True,
)
